/*************************************************************************\
**  source       * ha_soliddb_backup.h
**  directory    * mysql
**  description  * Function interface of Online Backup
**               * of solidDB Storage Engine for MySQL 5.1.
**               * 
**               * Copyright (C) 2007 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifndef HA_SOLIDDB_BACKUP_H
#define HA_SOLIDDB_BACKUP_H

bool schema_table_store_record(THD* thd, TABLE* table);

#ifdef __cplusplus
extern "C" {
#endif

#include <su0error.h>
#include <rs0sysi.h>
#include <su0cfgst.h>
#include <dbe0db.h>

#ifdef __cplusplus
}
#endif

#ifndef MYSQL_THD
#define MYSQL_THD THD*
#endif

/*************************************************************************\
 *                  Interface for DDL Monitoring
\*************************************************************************/

void  init_ddl_monitor();
void  free_ddl_monitor();
void  block_solid_DDL();
void  unblock_solid_DDL();
bool  is_solid_DDL_blocked();
bool  start_solid_DDL(int &rc);
void  finish_solid_DDL();

/*************************************************************************\
 *                  Interface for Backup Directory
\*************************************************************************/

#if MYSQL_VERSION_ID >= 50100
// ha_soliddb.cc requires access to this variable
extern char* soliddb_backupdir;
#endif

bool  init_backup_directory(su_inifile_t*);
#if MYSQL_VERSION_ID >= 50100
int   soliddb_check_backupdir(MYSQL_THD, struct st_mysql_sys_var*, void*, struct st_mysql_value*);
void  soliddb_update_backupdir(MYSQL_THD, struct st_mysql_sys_var*, void*, const void*);
#else
int   soliddb_check_backupdir(MYSQL_THD, set_var*);
bool  soliddb_update_backupdir(MYSQL_THD, set_var*);
#endif

/*************************************************************************\
 *                  Interface for Online Backup
\*************************************************************************/

void      init_backup();
void      free_backup();
su_ret_t  do_backup(rs_sysi_t*, char* parameters, su_err_t**);

#if MYSQL_VERSION_ID >= 50100
/*************************************************************************\
 *            Interface for Information Schema Extension
\*************************************************************************/

int soliddb_init_backup_status(void*);

#endif /* MYSQL_VERSION_ID >= 50100 */

#endif /* HA_SOLIDDB_IS_H */
