/*************************************************************************\
**  source       * ssdirw32.c
**  directory    * ss
**  description  * Directory handling routines for W32
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------

Implementation of directory routines for Win32

Limitations:
-----------


Error handling:
--------------


Objects used:
------------


Preconditions:
-------------


Multithread considerations:
--------------------------


Example:
-------


**************************************************************************
#endif /* DOCUMENTATION */
#include "ssenv.h"

#ifdef SS_NT /* Whole file */

#include "ssstdio.h"
#include "ssstring.h"

#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <stdlib.h>
#include <sys/locking.h>
#include <sys/utime.h>
#include <io.h>
#include "sschcvt.h"

#define DIRMASK         "\\*.*"
#define DIRATTR         (0x16)


#include "ssc.h"
#include "ssdir.h"
#include "ssmem.h"
#include "ssdebug.h"

#ifdef DOS
#define MAXNAMLEN   13

typedef struct {
        unsigned char   stuff[30];
        char            namebuffer[MAXNAMLEN];
} DIR;
#endif

struct SsDirStruct {

        SsDirEntT           d_dirent;
#ifdef MSVC7_NT
        intptr_t            d_handle;
#else
        long                d_handle;
#endif
        struct _finddata_t  d_fileinfo;

};


/*##**********************************************************************\
 * 
 *		SsDirOpen
 * 
 * Opens a directory for reading
 * 
 * Parameters : 
 * 
 *	pathname - in, use
 *		directory to be opened
 *
 * Return value - give :
 *
 *      pointer to directory structure
 * 
 * Limitations  :
 * 
 * Globals used :
 */
SsDirT *SsDirOpen(
        char *pathname)
{
        SsDirT* dir = SSMEM_NEW(SsDirT);
        size_t  pathlen = strlen(pathname);
        char*   path = SsMemAlloc(pathlen + strlen(DIRMASK) + 2);
        char*   dirmask = DIRMASK;

        strcpy(path, pathname);

        if (pathlen && path[pathlen-1] == dirmask[0]) {
            path[pathlen-1] = '\0';
        }
        strcat(path, DIRMASK);

        dir->d_handle = _findfirst(path, &dir->d_fileinfo);
        if (dir->d_handle == -1) {
            ss_dprintf_1(("_findfirst errno=%d.\n", errno));
            SsMemFree(dir);
            SsMemFree(path);
            return(NULL);
        }
        dir->d_dirent.filename = dir->d_fileinfo.name;
        dir->d_dirent.used = FALSE;
        SsMemFree(path);
        return(dir);
}


/*##**********************************************************************\
 * 
 *		SsDirRead
 * 
 * Reads the next file from a directory opened by SsDirOpen
 * 
 * Parameters : 
 * 
 *	dir - in out, use
 *		directory to be read
 *
 * 
 * Return value - ref : 
 * 
 *      pointer to a directory entry
 * 
 * Limitations  :
 * 
 * Globals used :
 */
SsDirEntT* SsDirRead(
        SsDirT *dir)
{
        if (dir->d_dirent.used) {

	    dir->d_dirent.filename = NULL;

            if (_findnext(dir->d_handle, &dir->d_fileinfo) != 0) {
                ss_dprintf_1(("_findnext errno=%d.\n", errno));
                return(NULL);
            }

            dir->d_dirent.filename = dir->d_fileinfo.name;
            dir->d_dirent.used = FALSE;
        }

        dir->d_dirent.used = TRUE;
        return(&(dir->d_dirent));
}

/*##**********************************************************************\
 * 
 *		SsDirClose
 * 
 * Closes a directory opened by SsDirOpen
 * 
 * Parameters : 
 * 
 *	dir - in, take
 *		directory to be closed
 *
 * Return value :
 * 
 * Limitations  :
 * 
 * Globals used :
 */
void SsDirClose(
        SsDirT *dir)
{
        _findclose(dir->d_handle);
        SsMemFree(dir);
}


#endif /* SS_NT */
