/*************************************************************************\
**  source       * ssproces.c
**  directory    * ss
**  description  * Portable process handling functions.
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------


Limitations:
-----------


Error handling:
--------------


Objects used:
------------


Preconditions:
-------------


Multithread considerations:
--------------------------


Example:
-------


**************************************************************************
#endif /* DOCUMENTATION */

#include "ssenv.h"

#if defined(SS_IRIX)
#  include <sys/types.h>
#  include <unistd.h>       /* prototype: fork() */
#elif defined(SS_UNIX)
#  include <unistd.h>
#endif
#if defined(SS_OSE) && defined(SS_ARM)
  /* These are defined in unistd.h and redefined in
     armcc's stdio.h. armcc is not very happy about that. */
#undef SEEK_SET
#undef SEEK_CUR
#undef SEEK_END
#endif

#include "ssstdio.h"
#include "ssstdlib.h"
#include "ssctype.h"
#include "sschcvt.h"

#include "ssc.h"
#include "ssmem.h"
#include "ssproces.h"

/***********************************************************************\
 ***                                                                 ***
 ***                             VMS                                 ***
 ***                                                                 ***
\***********************************************************************/

#if defined(SS_VMS)

#include <starlet.h>

/*##**********************************************************************\
 * 
 *		SsProcessSetPriority
 * 
 * Sets process' priority.
 * 
 * Parameters : 
 * 
 *	new_priority - in
 *		New priority. Actual value depends on operating system.
 *		
 *	p_old_priority - out
 *		Old priority.
 *		
 * Return value : 
 * 
 *      TRUE    - setting was succesfull
 *      FALSE   - setting failed
 * 
 * Comments : 
 * 
 * Globals used : 
 * 
 * See also : 
 */
bool SsProcessSetPriority(int new_priority, int* p_old_priority)
{
        long int stat;

        stat = sys$setpri(NULL, NULL, new_priority, p_old_priority);

        return(stat & 1);
}

/*##**********************************************************************\
 * 
 *		SsProcessSetWorkingSetSize
 * 
 * Sets working set memory size in bytes of the current process.
 * 
 * Parameters : 
 * 
 *	working_set - in
 *		New working set size in bytes.
 *		
 * Return value : 
 * 
 *      TRUE    - setting was succesfull
 *      FALSE   - setting failed
 * 
 * Comments : 
 * 
 * Globals used : 
 * 
 * See also : 
 */
bool SsProcessSetWorkingSetSize(long working_set_size)
{
        long int stat;
        long current_size;

        /* Get current working set size. */
        stat = sys$adjwsl(0, &current_size);
        if (!(stat & 1)) {
            return(FALSE);
        }

        /* Adjust the new working set size. First parameter gives the
           amount of pages to change from the current value. */
        working_set_size /= 512;
        sys$adjwsl(working_set_size - current_size, NULL);

        return(stat & 1);
}

void SsProcessSwitch(void)
{

}

long SsProcessId(void)
{
        return(0L);
}

#endif /* SS_VMS */


#if defined(SS_OS2) || defined(SS_DOS) || defined(SS_NT) || defined(SS_UNIX) || defined(SS_NLM) || defined(SS_E32) || defined(SS_WCE)

bool SsProcessSetPriority(int priority, int* p_prev_priority)
{
        SS_NOTUSED(priority);
        SS_NOTUSED(p_prev_priority);
        return(TRUE);
}

bool SsProcessSetWorkingSetSize(long working_set)
{
        SS_NOTUSED(working_set);
        return(TRUE);
}

void SsProcessSwitch(void)
{

}

#if (defined(SS_UNIX) || defined(SS_NT)) && !defined(SS_CRX) && !defined(SS_VXW) && !defined(SS_OSE)

long SsProcessId(void)
{
        return(getpid());
}

#else /* SS_UNIX || SS_NT */


#if defined(SS_WCE)

long SsProcessId(void)
{
        return( GetCurrentProcessId() );
}

#else

long SsProcessId(void)
{
        return(0L);
}
#endif /* SS_WCE */
#endif /* SS_UNIX || SS_NT */

#endif /* OS2, DOS, NT, UNIX */

/***********************************************************************\
 ***                                                                 ***
 ***                     WINDOWS                                     ***
 ***                                                                 ***
\***********************************************************************/

#if defined(SS_WIN)

#include "sswindow.h"
#include "ssdebug.h"

bool SsProcessSetPriority(int priority, int* p_prev_priority)
{
        SS_NOTUSED(priority);
        SS_NOTUSED(p_prev_priority);
        return(TRUE);
}

bool SsProcessSetWorkingSetSize(long working_set)
{
        SS_NOTUSED(working_set);
        return(TRUE);
}

/*##**********************************************************************\
 * 
 *		SsProcessSwitch
 * 
 * Function for applications that do not have an own messageloop.
 * Its main goal is to yield and thus give time for other applications.
 * 
 * Parameters : 	 - none
 * 
 * Return value : 
 * 
 * Comments : 
 * 
 * Globals used : 
 * 
 * See also : 
 */
void SsProcessSwitch(void)
{
        MSG msg;
        bool b = TRUE;

        ss_dprintf_1(("SsProcessSwitch.\n"));
        /* Just Yield */
        b = PeekMessage(&msg, 0, 0, 0, PM_REMOVE);
        if (b) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
#if defined(MSC) && !defined(SS_W32)
            static BOOL bSwitch = TRUE;
            if (bSwitch) {
                _asm mov ax, 0x1689
                _asm mov bl, 1; set mouse busy flag
                _asm int 0x2f
            } else {
                _asm int 0x28
            }
            bSwitch = !bSwitch;
#endif /* defined(MSC) && !defined(SS_W32) */
        }
}

long SsProcessId(void)
{
        return(0L);
}

#endif /* SS_WIN */

/***********************************************************************\
 ***                                                                 ***
 ***                     GENERAL                                     ***
 ***                                                                 ***
\***********************************************************************/

#if defined(SS_NT) || defined(SS_OS2) || defined(SS_UNIX)

static void make_newcmd(
        char* cmd,
        char** p_cmdname,
        char** p_cmdargs)
{
        char* newcmd;
        char* newargs;
        
        newcmd = SsMemStrdup(cmd);
        newargs = SsMemStrdup(cmd);
        *p_cmdname = newcmd;

        while (*newcmd != '\0' && !ss_isspace((ss_byte_t)*newcmd)) {
            newcmd++;
            newargs++;
        }
        if (*newcmd != '\0') {
            *newcmd = '\0';
            newcmd++;
            newargs--;
            *newargs = '\0';
            newargs++;
            /*
            while (*newcmd != '\0' && !ss_isspace((ss_byte_t)*newcmd)) {
                newcmd++;
            }
            */
            *p_cmdargs = newargs;
        } else {
            *p_cmdargs = NULL;
        }
}

#endif /* SS_NT or SS_OS2 or SS_UNIX */

/*##**********************************************************************\
 * 
 *		SsSystem
 * 
 * Replacement for C-libary system() function. Executes the given
 * command line.
 * 
 * Parameters : 
 * 
 *	cmd - 
 *		
 *		
 * Return value : 
 * 
 * Comments : 
 * 
 * Globals used : 
 * 
 * See also : 
 */
bool SsSystem(char* cmd)
{
#if defined(SS_NT) || defined(SS_OS2)

#ifdef MSVC7_NT
        uintptr_t status;
#else
        int status;
#endif
        char* cmdname;
        char* cmdargs;

        make_newcmd(cmd, &cmdname, &cmdargs);

        status = spawnlp(P_NOWAIT, cmdname, cmdargs, NULL);

        SsMemFree(cmdname);

        return(status == 0);

#elif defined(SS_UNIX) && !defined(SS_VXW) && !defined(SS_CRX)  && !defined(SS_OSE)

        int pid;
        char* cmdname;
        char* cmdargs;

        make_newcmd(cmd, &cmdname, &cmdargs);

        pid = fork();

        if (pid == 0) {
            execlp(cmdname, cmdargs, NULL);
            _exit (127);
        } else {
            SsMemFree(cmdname);
            SsMemFree(cmdargs);
        }

        return(pid > 0);

#else
        SS_NOTUSED(cmd);

        return(FALSE);

#endif
}


bool SsSystemSync(char* cmd)
{
#if defined(SS_NT) || defined(SS_OS2)

#ifdef MSVC7_NT
        intptr_t pid;
#else
        int pid;
#endif
        int status;
        char* cmdname;
        char* cmdargs;

        make_newcmd(cmd, &cmdname, &cmdargs);

        pid = spawnlp(P_NOWAIT, cmdname, cmdargs, NULL);

        SsMemFree(cmdname);
#if defined(GCC) && defined(SS_OS2)
        waitpid(pid, &status, 0);
#else /* GCC for OS/2 */
        cwait(&status, pid, WAIT_CHILD);
#endif /* GCC for OS/2 */
        return(status == 0);

#elif defined(SS_UNIX) && !defined(SS_VXW) && !defined(SS_CRX) && !defined(SS_OSE)

        int pid;
        char* cmdname;
        char* cmdargs;

        make_newcmd(cmd, &cmdname, &cmdargs);

        pid = fork();

        if (pid == 0) {
            execlp(cmdname, cmdargs, NULL);
            _exit (127);
        } else {
            SsMemFree(cmdname);
            SsMemFree(cmdargs);
        }

        return(pid > 0);

#elif defined(SS_VMS) && 0

        int status = 0, ret;
        char* line = NULL;
        char* cliline;
        struct {
            int len;
            char* ptr;
        } vms_string;

        line = cmd;
        if (!*line) {
            return(0); /* the line was empty */
        }
        cliline = SsMemAlloc(2 + strlen(line) + 1);
        strcpy(cliline, "$ ");
        strcat(cliline, line);
        vms_string.len = strlen(cliline);
        vms_string.ptr = cliline;
        ret = lib$spawn(
            &vms_string,
            0,      /* No output file */
            0,      /* No input file */
            0,      /* Flags: Create a subprocess and wait for completion */
            0,      /* No process name specified */
            0,      /* No process ID */
            &status,/* Completion status */
            0,      /* No completion event flag */
            0,      /* No AST routine */
            0,      /* No storage for AST routine */
            0,      /* No prompt string */
            0);     /* No CLI specified */
        SsMemFree(cliline);

        if ((ret&1) != 1) {

            /* Delete the output file that the command was supposed to make.  */
            struct stat st;
            printf("Error %d\n", status);

            if (signal_delete_file
	        && stat (signal_delete_file, &st) >= 0
	        && (st.st_mode & S_IFMT) != S_IFDIR
	        && st.st_mtime != signal_delete_mtime) {

	        SsFPrintf (SsStderr, "\nDeleting file %s\n", signal_delete_file);
	        SsFFlush (SsStderr);
	        unlink (signal_delete_file);
	    }
        } else {
            status = 0;
        }
        if (tmp != NULL) {
            SsMemFree(tmp);
        }
        return(status == 0);

#else
        SS_NOTUSED(cmd);

        return(FALSE);

#endif
}
